pragma solidity >=0.5.11;

/*
 * This system is intended to reflect the following rule on prescriptions:
 *    A prescription cannot be filled more times than the designated number of refills specifies.
 *
 * However, the Drug Enforcement Agency is worried that the code below might allow violations of the rule by
 * malevolent patients, who might modify or subclass the Patient contract.
 * Specifically, it is possible for a caller to call depositPrescription() several times
 * on different Pharmacy objects with the same Prescription object. See getMedicine() below for an example.
 *
 * YOUR TASK: Please use what you have learned today to fix this problem (avoiding runtime checks if possible).
 */

// Prescriptions are immutable by design; pharmacies that want to record information about
// prescriptions must do so in a separate data structure.
// Only approved doctors can create new prescriptions.
contract Prescription {
    string public prescriptionText; // specifies the drug, dosage, etc.
    int public refills;
    string public patientName; // Assume that patient names are unique.

    constructor (string memory text, int r, string memory patient) public {
        // Assume there is code to prevent unauthorized creation of prescriptions.
        prescriptionText = text;
        refills = r;
        patientName = patient;
    }
}

contract PrescriptionRecordsProvider {
    int nextID;

    struct PharmacyPrescriptionRecord {
        int fillsLeft;
        Prescription prescription;
    }
    mapping(int => PharmacyPrescriptionRecord) prescriptionRecords;

    constructor() public {
        nextID = 0;
    }

	// Must be called before the patient can fill a new prescription.
    // Assumes that the prescription was not already deposited in this Pharmacy.
    // Consumes ownership of prescription
	function depositPrescription(Prescription prescription) public returns (int) {
        // Look for duplicates
        for (int i = 0; i < nextID; i++) {
            Prescription record = prescriptionRecords[i].prescription;

            if (record == prescription) {
                revert("Can't deposit the same prescription twice");
            }
        }

        int id = nextID;
        int origFills = prescription.refills();
        addRecord(id, PharmacyPrescriptionRecord(origFills, prescription));
        nextID++;
	}

    // If the prescription has fills remaining,
    // this function decrements the fill count
    // and returns true
    // Otherwise does nothing and returns false
    function fill(int id) public returns (bool) {
        PharmacyPrescriptionRecord memory record = removeRecord(id);
        if (record.fillsLeft > 0) {
            record.fillsLeft--;

            addRecord(id, record);
            return true;
        }
        addRecord(id, record);
        return false;
    }

    function removeRecord(int prescriptionID) private returns (PharmacyPrescriptionRecord memory) {
        PharmacyPrescriptionRecord memory record = prescriptionRecords[prescriptionID];
        delete prescriptionRecords[prescriptionID];
        return record;
    }

    function addRecord(int prescriptionID, PharmacyPrescriptionRecord memory record) private {
        prescriptionRecords[prescriptionID] = record;
    }
}

/* A Pharmacy keeps a record of all the prescriptions that patients have submitted as well as how many times each prescription has been filled.
 * Before filling a prescription, the pharmacy must make sure that there is at least one refill available.
 */
contract Pharmacy {
    PrescriptionRecordsProvider private provider; // shared

    // requires shared reference to a PrescriptionRecordsProvider
    constructor(PrescriptionRecordsProvider theProvider) public {
        provider = theProvider;
    }

    function depositPrescription(Prescription prescription) public returns (int) {
        return provider.depositPrescription(prescription);
    }

	function fillPrescription(int prescriptionID) public {
        if (provider.fill(prescriptionID)) {
            requestFillFromPharmacist(prescriptionID);
        }
        else {
            revert("No fills remaining!");
        }
	}

	function requestFillFromPharmacist(int prescriptionID) public {
		// This code, which is not shown, notifies the pharmacist to actually fill the prescription.
	}
}

// Do not change the code below here. Instead, change the code ABOVE to defend against the code below.
contract Patient {
    function getMedicine(Pharmacy pharmacy1, Pharmacy pharmacy2, Prescription prescription) public {
        int prescriptionID1 = pharmacy1.depositPrescription(prescription);
        pharmacy1.fillPrescription(prescriptionID1);

        // Uh oh: the patient should not be able to deposit the prescription
        // a second time, obtaining more refills than authorized!
        int prescriptionID2 = pharmacy2.depositPrescription(prescription);
        pharmacy2.fillPrescription(prescriptionID2);
    }
}

